require 'spec_helper'

describe SidekiqStrategies::AccountHandler do

  before(:each) do
    @subscriber = create_double_subscriber
    @acc_handler = SidekiqStrategies::AccountHandler.new(120, 250, :query, 0, Spool::Publisher)
    @acc_handler.add_subscriber(@subscriber)
  end

  describe '#new' do

    it 'return an instance of AccountHandler object' do
      expect( @acc_handler ).to be_an(SidekiqStrategies::AccountHandler)
    end

    it 'should have an Account reference with id 120' do
      expect( @acc_handler.account ).to be_an(Account)
      expect( @acc_handler.account.id ).to be(120)
    end

    it 'should have a Transaction reference with id 250' do
      expect( @acc_handler.transaction ).to be_an(Transaction)
      expect( @acc_handler.transaction.id ).to be(250)
    end

  end

  describe 'instance methods' do

    describe '#save_sidekiq_jid' do
      it 'should save jid value' do
        @acc_handler.save_sidekiq_jid('jid123')
        expect( @acc_handler.transaction.sidekiq_jid ).to eql('jid123')
      end

      it 'should call Transaction.save! with no params' do
        expect( @acc_handler.transaction ).to receive(:save!)
        @acc_handler.save_sidekiq_jid('jid123')
      end
    end

    describe '#save_beanstalkd_jid' do
      it 'should save jid value' do
        @acc_handler.save_beanstalkd_jid(562)
        expect( @acc_handler.transaction.beanstalkd_jid ).to eql(562)
      end

      it 'should call Transaction.save! with no params' do
        expect( @acc_handler.transaction ).to receive(:save!)
        @acc_handler.save_beanstalkd_jid(562)
      end
    end

    describe '#success' do
      it 'should set account state to ready' do
        @acc_handler.success
        expect( @acc_handler.account.state ).to eql('ready')
      end

      it 'should notify subscribers' do
        expect( @subscriber ).to receive(:state_changed).with(@acc_handler.account, 'reserved', 'ready')
        @acc_handler.success
      end
    end

    describe '#error' do
      it 'should set account state to error' do
        @acc_handler.error
        expect( @acc_handler.account.state ).to eql('error')
      end

      it 'should notify subscribers' do
        expect( @subscriber ).to receive(:state_changed).with(@acc_handler.account, 'reserved', 'error')
        @acc_handler.error
      end
    end

    describe '#start_action' do
      it 'should set account state to querying' do
        @acc_handler.start_action
        expect( @acc_handler.account.state ).to eql('querying')
      end

      it 'should notify subscribers' do
        expect( @subscriber ).to receive(:state_changed).with(@acc_handler.account, 'reserved', 'querying')
        @acc_handler.start_action
      end
    end

    describe '#queue_for_retry' do
      it 'should set account state to awaiting_retry' do
        @acc_handler.queue_for_retry
        expect( @acc_handler.account.state ).to eql('awaiting_retry')
      end

      it 'should increase retry_count in 1' do
        begin_retry_count = @acc_handler.transaction.retry_count
        @acc_handler.queue_for_retry
        expect( @acc_handler.transaction.retry_count ).to be(begin_retry_count + 1)
      end

      it 'should call Transaction.save! with params "validate: false"' do
        expect( @acc_handler.transaction ).to receive(:save!).with(validate: false)
        @acc_handler.queue_for_retry
      end

      it 'should notify subscribers' do
        expect( @subscriber ).to receive(:state_changed).with(@acc_handler.account, 'reserved', 'awaiting_retry')
        @acc_handler.queue_for_retry
      end
    end

    describe '#queue_for_wait_at_worker' do
      it 'should set account state to waiting_at_worker' do
        @acc_handler.queue_for_wait_at_worker
        expect( @acc_handler.account.state ).to eql('waiting_at_worker')
      end

      it 'should notify subscribers' do
        expect( @subscriber ).to receive(:state_changed).with(@acc_handler.account, 'reserved', 'waiting_at_worker')
        @acc_handler.queue_for_wait_at_worker
      end
    end

    describe '#save_success_response' do
      it 'should update Transaction.resonse' do
        @acc_handler.save_success_response(create_worker_response)
        expect( @acc_handler.transaction.response ).to eq(create_worker_response)
      end

      it 'should update Transaction.amount' do
        @acc_handler.save_success_response(create_worker_response)
        expect( @acc_handler.transaction.amount ).to eq(create_worker_response[:amount_due])
      end

      it 'should update account response' do
        @acc_handler.save_success_response(create_worker_response)
        expect( @acc_handler.account.response ).to_not be(nil)
      end

      it 'should call Transaction.save! with params "validate: false"' do
        expect( @acc_handler.transaction ).to receive(:save!).with(validate: false)
        @acc_handler.queue_for_retry
      end
    end

    describe '#save_exception' do
      before(:each) do
        @exception = XBP::Error::ApiInternalError.new('test error')
        @acc_handler.save_exception(@exception)
      end

      it 'shuold update Transaction.response' do
        expect( @acc_handler.transaction.response ).to_not be(nil)
      end

      it 'shuold update Transaction.response_code' do
        exception = XBP::Error::ApiInternalError.new('test error')
      end

      it 'shuold update Transaction.response_messsage' do
        expect( @acc_handler.transaction.response_message ).to_not be(nil)
      end

      it 'shuold update Transaction.exception_type' do
        expect( @acc_handler.transaction.exception_type ).to eql("XBP::Error::ApiInternalError")
      end

      it 'shuold update Transaction.exception_message' do
        expect( @acc_handler.transaction.exception_message ).to eql("test error")
      end

      it 'should call Transaction.save! with params "validate: false"' do
        expect( @acc_handler.transaction ).to receive(:save!).with(validate: false)
        @acc_handler.save_exception(@exception)
      end
    end

  end
  
end