require 'spec_helper'

describe SidekiqStrategies::PollingStrategy do

  before(:each) do
    @logger             = create_double_logger
    @airbrake_notifier  = create_double_airbrake
    
    @acc_handler  = SidekiqStrategies::AccountHandler.new(120, 250, :query, 0, Spool::Publisher)
    @helper       = SidekiqStrategies::Helper.new(@acc_handler, Rails.application.routes.url_helpers)
    @strategy     = SidekiqStrategies::PollingStrategy.new(nil, nil, nil, nil, nil)
  end

  describe '#new' do

    it 'return an instance of PollingStrategy object' do
      expect( @strategy ).to be_an(SidekiqStrategies::PollingStrategy)
    end

    it 'should respond to #perform' do
      expect( @strategy.respond_to?('perform') ).to be(true)
    end

    it 'should respond to #retries_exhausted' do
      expect( @strategy.respond_to?('retries_exhausted') ).to be(true)
    end

    it 'should expect 4 params on constructor' do
      expect{
        SidekiqStrategies::PollingStrategy.new
      }.to raise_error

      expect{
        SidekiqStrategies::PollingStrategy.new(nil, nil, nil, nil, nil)
      }.to_not raise_error
    end

  end

  describe '#perform' do
    
    context 'when Beanstalkd service is not reachable' do
      before(:each) do
        pool = create_double_beaneater_pool
        allow(pool).to receive(:tubes).and_raise(Beaneater::NotConnected)

        @bns_handler  = SidekiqStrategies::BeanstalkdHandler.new(pool)
        @strategy     = SidekiqStrategies::PollingStrategy.new(@acc_handler, @bns_handler, @logger, @airbrake_notifier, @helper)
      end 

      it 'should raise XBP::Error::MessageQueueNotReachable error' do
        expect{
          @strategy.perform('abc123')
        }.to raise_error(XBP::Error::MessageQueueNotReachable)
      end

      it 'should set account state to awaiting_retry' do
        begin
          @strategy.perform('abc123')
        rescue XBP::Error::MessageQueueNotReachable=>e
          expect( @acc_handler.account.state ).to eql('awaiting_retry')
        end
      end
    end

    context 'when there are no workers available' do
      before(:each) do
        request_tube_stats  = create_stats({:current_watching=>0})
        request_tube        = create_double_beaneater_tube(:name=>'pronet', :stats=>request_tube_stats)
        tubes               = {'pronet'=>request_tube}
        pool                = create_double_beaneater_pool({:tubes=>tubes})

        @bns_handler  = SidekiqStrategies::BeanstalkdHandler.new(pool)
        @strategy     = SidekiqStrategies::PollingStrategy.new(@acc_handler, @bns_handler, @logger, @airbrake_notifier, @helper)
      end 

      it 'should raise XBP::Error::WorkersUnavailable error' do
        expect{
          @strategy.perform('abc123')
        }.to raise_error(XBP::Error::WorkersUnavailable)
      end

      it 'should set account state to awaiting_retry' do
        begin
          @strategy.perform('abc123')
        rescue XBP::Error::WorkersUnavailable=>e
          expect( @acc_handler.account.state ).to eql('awaiting_retry')
        end
      end
    end

    context 'when account state is not included in [reserved, awaiting_retry, waiting_at_worker]' do
      context 'when there is no previous data' do
        before(:each) do
          @acc_handler.account.state = 'querying'

          request_tube  = create_double_beaneater_tube(:name=>'pronet')
          tubes         = {'pronet'=>request_tube}
          pool          = create_double_beaneater_pool({:tubes=>tubes})

          @bns_handler  = SidekiqStrategies::BeanstalkdHandler.new(pool)
          @strategy     = SidekiqStrategies::PollingStrategy.new(@acc_handler, @bns_handler, @logger, @airbrake_notifier, @helper)
        end 

        it 'should not raise error' do
          expect{
            @strategy.perform('abc123')
          }.to_not raise_error
        end

        it 'should set account state to error' do
          @strategy.perform('abc123')
          expect( @acc_handler.account.state ).to eql('error')
        end
      end
    end

    context 'when response is never received' do
      before(:each) do
        request_tube_stats  = create_stats({:current_watching=>1})
        reply_tube_stats    = create_stats({:state=>'ready'})
        request_tube        = create_double_beaneater_tube(:name=>'pronet', :stats=>request_tube_stats, :put_response=>{:id=>1})
        reply_tube          = create_double_beaneater_tube({:name=>'gt-120'})
        tubes               = {'pronet'=>request_tube, 'gt-120'=>reply_tube}
        request_job_stats   = create_stats({:state=>'ready'})
        request_job         = create_double_beaneater_job({:stats=>request_job_stats})
        jobs                = create_double_beaneater_jobs({1=>request_job})
        pool                = create_double_beaneater_pool({:tubes=>tubes, :jobs=>jobs})

        allow(reply_tube).to receive(:reserve).with(1).and_raise(Beaneater::TimedOutError.new(nil, nil))

        @bns_handler  = SidekiqStrategies::BeanstalkdHandler.new(pool)
        @strategy     = SidekiqStrategies::PollingStrategy.new(@acc_handler, @bns_handler, @logger, @airbrake_notifier, @helper)

        @strategy.poll_tries_before_airbrake  = 10
        @strategy.poll_max_tries              = 20
      end

      it 'should set account state to error' do
        @strategy.perform('abc123')
        expect( @acc_handler.account.state ).to eql('error')
      end

      it 'should send an airbrake when 10 seconds of polling has been reached' do
        expect( @airbrake_notifier ).to receive(:notify).with(:error_message => "Polling is taking more than 10 seconds.")
        @strategy.perform('abc123')
      end

      it 'should send an airbrake when 20 seconds of polling has been reached' do
        expect( @airbrake_notifier ).to receive(:notify).with(:error_message => "Queue gt-120 time out. Waited 20 seconds.")
        @strategy.perform('abc123')
      end
    end

    context 'when response is received' do
      context 'when is success' do
        before(:each) do
          request_tube_stats  = create_stats({:current_watching=>1})
          request_tube        = create_double_beaneater_tube(:name=>'pronet', :stats=>request_tube_stats, :put_response=>{:id=>1})
          reply_tube          = create_double_beaneater_tube({:name=>'gt-120'})
          tubes               = {'pronet'=>request_tube, 'gt-120'=>reply_tube}
          request_job_stats   = create_stats({:state=>'ready'})
          request_job         = create_double_beaneater_job({:stats=>request_job_stats})
          reply_job           = create_double_beaneater_job({:body=>Oj.dump(create_worker_response)})
          jobs                = create_double_beaneater_jobs({1=>request_job})
          pool                = create_double_beaneater_pool({:tubes=>tubes, :jobs=>jobs})

          allow(reply_tube).to receive(:reserve).with(1).and_return(reply_job)          

          @bns_handler  = SidekiqStrategies::BeanstalkdHandler.new(pool)
          @strategy     = SidekiqStrategies::PollingStrategy.new(@acc_handler, @bns_handler, @logger, @airbrake_notifier, @helper)
        end

        it 'should set account state to ready' do
          @strategy.perform('abc123')
          expect( @acc_handler.account.state ).to eql('ready')
        end
      end

      context 'when is error' do
        before(:each) do
          request_tube_stats  = create_stats({:current_watching=>1})
          request_tube        = create_double_beaneater_tube(:name=>'pronet', :stats=>request_tube_stats, :put_response=>{:id=>1})
          reply_tube          = create_double_beaneater_tube({:name=>'gt-120'})
          tubes               = {'pronet'=>request_tube, 'gt-120'=>reply_tube}
          request_job_stats   = create_stats({:state=>'ready'})
          request_job         = create_double_beaneater_job({:stats=>request_job_stats})
          reply_job           = create_double_beaneater_job({:body=>Oj.dump({:exception=>create_worker_exception})})
          jobs                = create_double_beaneater_jobs({1=>request_job})
          pool                = create_double_beaneater_pool({:tubes=>tubes, :jobs=>jobs})

          allow(reply_tube).to receive(:reserve).with(1).and_return(reply_job)          

          @bns_handler  = SidekiqStrategies::BeanstalkdHandler.new(pool)
          @strategy     = SidekiqStrategies::PollingStrategy.new(@acc_handler, @bns_handler, @logger, @airbrake_notifier, @helper)
        end

        it 'should set account state to error' do
          @strategy.perform('abc123')
          expect( @acc_handler.account.state ).to eql('error')
        end
      end
    end
  end

end